--  Awesome Tower Defense
--  Copyright (C) 2008-2009  Trygve Vea and contributors (read AUTHORS)
--  Copyright (C) 2010 Sean Bolton
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License along
--  with this program; if not, write to the Free Software Foundation, Inc.,
--  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

-- globals
gamecycle = 0
game_paused = false
lastkeys = ""

function event_loop()
    local eventqueue -- SDL_Event
    local last_render = -20
    local last_cycle = SDL.GetTicks()
    local last_fps = last_cycle
    local cycle_speed = 10

    while true do
        local top = SDL.GetTicks()
        eventqueue = SDL.PollEvent()

        while eventqueue do
            if eventqueue.type == SDL.QUIT then
                return

            elseif eventqueue.type == SDL.MOUSEMOTION then
                if current_screen == SCREEN_MENU then update_menu_selection(eventqueue.x,eventqueue.y) end
                if current_screen == SCREEN_INGAME then update_tooltip(eventqueue.x,eventqueue.y) end
                if current_screen == SCREEN_TUTORIAL then update_tutorial_selection(eventqueue.x,eventqueue.y) end
                update_cursor(eventqueue.x,eventqueue.y)

            elseif eventqueue.type == SDL.MOUSEBUTTONDOWN then
                if current_screen == SCREEN_MENU then
                    if click_menu() then return end
                    if current_screen == SCREEN_INGAME then
                        last_cycle = SDL.GetTicks()
                        cycle_speed = 10
                    end
                elseif current_screen == SCREEN_CREDITS then
                    show_menu()
                elseif current_screen == SCREEN_HIGHSCORES then
                    show_menu()
                elseif current_screen == SCREEN_TUTORIAL then
                    show_menu()
                elseif current_screen == SCREEN_INGAME then
                    local mx = math.floor(eventqueue.x/RECTSIZE_X)
                    local my = math.floor(eventqueue.y/RECTSIZE_Y)
                    if mx > 0 and mx < 19 and my > 0 and my < 13 then
                        if map:tower_allowed(mx, my) then
                            add_tower(mx, my, towerbutton)
                        elseif map:get_tower(mx, my) > 0 then
                            select_tower(mx, my)
                        end
                    elseif eventqueue.button == SDL.BUTTON_LEFT then
                        press_button(eventqueue.x, eventqueue.y)
                    end
                end

            elseif eventqueue.type == SDL.MOUSEBUTTONUP then
                --

            elseif eventqueue.type == SDL.KEYDOWN then
                if current_screen == SCREEN_MENU then
                    --

                elseif current_screen == SCREEN_INGAME then
                    if eventqueue.sym == SDL.K_ESCAPE then
                        show_menu()
                    elseif eventqueue.sym == SDL.K_1 then
                        press_buttonnumber(1)
                    elseif eventqueue.sym == SDL.K_2 then
                        press_buttonnumber(2)
                    elseif eventqueue.sym == SDL.K_3 then
                        press_buttonnumber(3)
                    elseif eventqueue.sym == SDL.K_q then
                        if selected_tower > 0 then tower_algorithm(selected_tower,1) end
                    elseif eventqueue.sym == SDL.K_w then
                        if selected_tower > 0 then tower_algorithm(selected_tower,2) end
                    elseif eventqueue.sym == SDL.K_e then
                        if selected_tower > 0 then tower_algorithm(selected_tower,3) end
                    elseif eventqueue.sym == SDL.K_r then
                        if selected_tower > 0 then tower_algorithm(selected_tower,4) end
                    elseif eventqueue.sym == SDL.K_t then
                        if selected_tower > 0 then tower_algorithm(selected_tower,5) end
                    elseif eventqueue.sym == SDL.K_s then
                        if selected_tower > 0 then sell_tower(selected_tower) end
                    elseif eventqueue.sym == SDL.K_p then
                        if selected_tower > 0 then toggle_shooting(selected_tower) end
                    elseif eventqueue.sym == SDL.K_f then
                        cycle_speed = 12 - cycle_speed
                    elseif eventqueue.sym == SDL.K_SPACE then
                        game_paused = not game_paused
                        if not games_paused then
                            last_cycle = SDL.GetTicks()
                            cycle_speed = 10
                        end
                    end

                    if eventqueue.sym < 256 then
                        lastkeys = string.sub(lastkeys .. string.char(eventqueue.sym), -5, -1)
                        if lastkeys == "iddqd" then iddqd() end
                        if lastkeys == "idkfa" then idkfa() end
                    end
                end

            elseif eventqueue.type == SDL.KEYUP then
                --

            end
            eventqueue = SDL.PollEvent()
        end

        -- if game is running, cycle it 100 times per second
        if current_screen == SCREEN_INGAME and not game_paused then
          while top - last_cycle > cycle_speed do
            move_projectile()
            move_monster()
            if (gamecycle % 20) == 0 then animate_sprites() end
            if (gamecycle % 10) == 0 then shoot_towers() end
            if (gamecycle % 100) == 0 then if spawn_monster() then return end end
            gamecycle = gamecycle + 1
            last_cycle = last_cycle + cycle_speed
          end
        end

        -- update frame-per-second display once per second
        if top - last_fps >= 1000 then
            getrenders()
            last_fps = last_fps + 1000
        end

        -- do render no more than 50 times per second
        if top - last_render >= 20 then
            render()
            last_render = top
        end

        -- shoot for starting the next loop 20ms after this one started
        local delay = (top + 20) - SDL.GetTicks()
        if delay > 0 then
            SDL.Delay(delay)
        end
    end
end

