--  Awesome Tower Defense
--  Copyright (C) 2008-2009  Trygve Vea and contributors (read AUTHORS)
--  Copyright (C) 2010 Sean Bolton
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License along
--  with this program; if not, write to the Free Software Foundation, Inc.,
--  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

-- Rules for level maps:

-- - The border of a map consists of its outermost characters: the top row,
--     the leftmost and rightmost columns, and the bottom two rows.
-- - The border of a map must be 'X's, with the following exceptions:
--    - 's' and 'e' for single-path maps, 's' is a monster start point,
--        'e' is a monster end point. There can be a maximum of 20 start
--        and end points each.
--    - 'S' and 'E' are reserved for double-path maps (like DTD
--        medium level), which is not implemented yet.
-- - The border is typically hidden by the game, so monsters appear to
--     originate from off-field and escape off-field
-- - To reiterate, only these characters are legal on the border of a map:
--      'X' - neither monsters nor towers allowed
--      's' - start, monster set 1, no tower allowed
--      'e' - end,   monster set 1, no tower allowed
--      'S' - start, monster set 2, no tower allowed
--      'E' - end,   monster set 2, no tower allowed

-- - Inside the map, only the following characters may be used:
--      '.' - towers and monsters okay       "path" (wide) 
--      '+' - towers okay, monsters not      "grass"
--      ' ' - monsters okay, towers not      "path" (narrow)
--      '*' - neither monsters nor towers    "rocks"
--    - The following are 'monsters okay, towers not' markers of direction
--        change for 'boing' pathfinding only:
--      '^' - causes monster to turn N
--      '>' - turn E
--      'v' - turn S
--      '<' - turn W
--      '/' - mirror S/W or N/E
--      '\' - mirror S/E or N/W
--      '(' - right turn
--      ')' - left turn
--      '?' - random left or right turn
--    - Start and end characters may also be placed inside the map, if you
--        don't mind the monster suddenly appearing or disappearing:
--      's' - start, monster set 1, no tower allowed
--      'e' - end,   monster set 1, no tower allowed
--      'S' - start, monster set 2, no tower allowed
--      'E' - end,   monster set 2, no tower allowed

-- The actual pathfinding code is in awesometd-ysdl.c, class td_map

levels = {
    { -- awesometd 0.3 Level 1!
        description = "The original awesometd 0.3 level 1",
        background = "background-level1.png",
        pathfinding = "boing",  -- currently must be 'boing' or 'lee'
        map = "XXXsXXXXXXXXXXXXXXXX" ..
              "X++ +++++++++++++++X" ..
              "X++ ++++++++++++>  e" ..
              "X++ +++v    <+++ ++X" ..
              "X++ +++ ++++ +++ ++X" ..
              "X++ +++ ++++ +++ ++X" ..
              "X++ +++ ++++ +++ ++X" ..
              "X++ +++>        ^++X" ..
              "X++ ++++++++ ++++++X" ..
              "X++ ++++++++ ++++++X" ..
              "X++ ++++++++ ++++++X" ..
              "X++>        ^++++++X" ..
              "X++++++++++++++++++X" ..
              "XXXXXXXXXXXXXXXXXXXX" ..
              "XXXXXXXXXXXXXXXXXXXX"
    },
    {
        -- demonstrate other 'boing' turns, plus auto-generate background
        description = "more 'boing' turns, autogen bg",
        pathfinding = "boing",
        map = "XXXsXXXXXXXXXXXXXXXX" ..
              "X++ +++++++++++++++X" ..
              "X++ ++>         v++X" ..
              "X++ ++ +++++++++ ++X" ..
              "X++ ++ +++++++++ ++X" ..
              "X++)            (++X" ..
              "X++ ++ +++++++++ ++X" ..
             [[X++ ++^   \++**+ ++X]] ..
             [[X++ ++ +++? \+++ ++X]] ..
             [[X++ ++\ \+ +? \+ ++X]] ..
             [[X++ ++++ + + +? /++X]] ..
             [[e  <+**+\ < < /++++X]] ..
              "X++++***+++++++++++X" ..
              "XXXXXXXXXXXXXXXXXXXX" ..
              "XXXXXXXXXXXXXXXXXXXX"
    },
    --{
    --    -- demonstrate other 'boing' turns, plus auto-generate background
    --    pathfinding = "boing",
    --    map = "XXXsXXXXXXXXXXXXXXXX" ..
    --          "X++ +++++++++++++++X" ..
    --         [[X++ ++/    \+++++++X]] ..
    --         [[X++ ++ ++++\    \++X]] ..
    --          "X++ ++ +**++++++ ++X" ..
    --          "X++ ++ +****+**+ ++X" ..
    --          "X++ ++ +++++++++ ++X" ..
    --          "X++)            (++X" ..
    --          "X++ ++ +++++++++ ++X" ..
    --          "X++ ++ +++++++++ ++X" ..
    --         [[X++ ++\         /++X]] ..
    --          "e  <+++++++++++++++X" ..
    --          "X++++++++++++++++++X" ..
    --          "XXXXXXXXXXXXXXXXXXXX" ..
    --          "XXXXXXXXXXXXXXXXXXXX"
    --},
    { -- Lee algorithm test 1
        description = "Lee pathfinding, BMP background",
        background = "background-lee_test_1.png",
        pathfinding = "lee",
        map = "XXXXXXXXXXXXXXXXXXXX" ..
              "X+++++++++++++++++*X" ..
              "s................++X" ..
              "s.................+X" ..
              "s.................+X" ..
              "X++++++++++++++...+X" ..
              "X+................+X" ..
              "X+....+++++++++++++X" ..
              "X+....+...........+X" ..
              "X+....+..+.........e" ..
              "X+....*..+.........e" ..
              "X++......+.........e" ..
              "X*+++++++++++++++++X" ..
              "XXXXXXXXXXXXXXXXXXXX" ..
              "XXXXXXXXXXXXXXXXXXXX"
    },
    { -- DTD silly thing
        description = "DTD-wanna-be (use idkfa)",
        background = "background-dtdx.png",
        pathfinding = "lee",
        map = "XXXXXXXXXXXXXXXXXXXX" ..
              "X******************X" ..
              "X******************X" ..
              "X**..............**X" ..
              "X**..............**X" ..
              "s  ..............**X" ..
              "s  ..............  e" ..
              "s  ..............  e" ..
              "X**..............  e" ..
              "X**..............**X" ..
              "X**..............**X" ..
              "X******************X" ..
              "X******************X" ..
              "XXXXXXXXXXXXXXXXXXXX" ..
              "XXXXXXXXXXXXXXXXXXXX"
    },
    { -- start and exit inside map
        description = "test: start and exit inside map",
        pathfinding = "lee",
        map = "XXXXXXXXXXXXXXXXXXXX" ..
              "X++++s++++++++++++*X" ..
              "X+s   ...........++X" ..
              "X+++..............+X" ..
              "X+s  .............+X" ..
              "X++++++++++++++...+X" ..
              "X+...           ..+X" ..
              "X+....+++++++++++++X" ..
              "X+....+.   ++++++++X" ..
              "X+....+..+ ++++++++X" ..
              "X+....*..+ ++++++++X" ..
              "X++..   .+    e++++X" ..
              "X*+++++++++++++++++X" ..
              "XXXXXXXXXXXXXXXXXXXX" ..
              "XXXXXXXXXXXXXXXXXXXX"
    },
    { -- diagonal test
        description = "diagonal test",
        pathfinding = "lee",
        map = "XXXsXXXXXXXXXXXXXXXX" ..
              "X++..+++++++......+X" ..
              "X++...+++++...+++.+X" ..
              "X+++...+++...++++.+X" ..
              "X++++...+...++++..+X" ..
              "X+++++.....++++...+X" ..
              "X++++++++++++++..++X" ..
              "X+++++.....++++.+++X" ..
              "X++++...+...+++..++X" ..
              "X+++...+++...++...+X" ..
              "X++...+++++...++..+X" ..
              "X+...+++++++...++.+X" ..
              "X+..+++++++++.....+X" ..
              "XXeXXXXXXXXXXXXXXXXX" ..
              "XXXXXXXXXXXXXXXXXXXX"
    },
    { -- test of entrances and exits on the border where they could interfere with other things
        description = "border entrance/exit test",
        pathfinding = "lee",
        map = "XXXXXXXXXXsXXXXXXXsX" ..
              "X+++++++++.+++++++.X" ..
              "X+++++++++.++++++..X" ..
              "X+++++++++..++++...X" ..
              "X+++++++++...++...+X" ..
              "X++++++++++..+...++X" ..
              "X+++++++++++....+++X" ..
              "X+++++.....+...++++X" ..
              "X++++...+.....+++++X" ..
              "X+++...+++...++++++X" ..
              "X++...+++++..++++++X" ..
              "X++..+++++++.++++++X" ..
              "X++..+++++++.++++++X" ..
              "XXXsXXXXXXXXeXXXXXXX" ..
              "XXXXXXXXXXXXXXXXXXXX"
    },
}

map = nil -- td_map

currentlevel = 1

tiles = nil

-- void load_level(SDL_Surface *bg, int lid) {
function load_level(bg, lid)
    local level = levels[lid]

    -- load map
    map, err = td.new_map(VIDEOMODE_WIDTH, VIDEOMODE_HEIGHT, RECTSIZE_X, RECTSIZE_Y,
                          level.map, level.pathfinding)
    if not map then
        print(string.format("Error: could not load map for level %d: %s", lid, err))
        -- and ... crash.
    end

    -- load or create background
    local tmpbg
    if level.background then
        tmpbg = load_surface_from_file(level.background)
    end
    if tmpbg then
        SDL.BlitSurface(tmpbg, nil, bg, nil)
    else
        local level_tiles = {}
        -- full-size: all one thing
        level_tiles['b'] = SDL.newRect(  0,  0, 32, 32)  -- Background
        level_tiles['g'] = SDL.newRect( 32,  0, 32, 32)  -- Grass
        level_tiles['r'] = SDL.newRect( 32, 32, 32, 32)  -- Rocks
        -- quarter-tile size: image depends on content of neighboring cells
        level_tiles[1] = SDL.newRect( 64,  0, 16, 16)  -- cell is path with grass on top and left
        level_tiles[2] = SDL.newRect( 80,  0, 16, 16)  -- top
        level_tiles[3] = SDL.newRect( 96,  0, 16, 16)  -- top-right
        level_tiles[4] = SDL.newRect( 64, 16, 16, 16)  --     left
        level_tiles[5] = SDL.newRect( 80, 16, 16, 16)  -- all path
        level_tiles[6] = SDL.newRect( 96, 16, 16, 16)  --     right
        level_tiles[7] = SDL.newRect( 64, 32, 16, 16)  -- bottom-left
        level_tiles[8] = SDL.newRect( 80, 32, 16, 16)  -- bottom
        level_tiles[9] = SDL.newRect( 96, 32, 16, 16)  -- bottom-right
        level_tiles[10] = SDL.newRect(  0, 32, 16, 16)  -- cell is path with grass only in bottom-right corner
        level_tiles[11] = SDL.newRect( 16, 32, 16, 16)  -- bottom-left
        level_tiles[12] = SDL.newRect(  0, 48, 16, 16)  -- top-right
        level_tiles[13] = SDL.newRect( 16, 48, 16, 16)  -- top-left
        if not tiles then
            tiles = load_surface_from_file("leveltiles.bmp"):DisplayFormat()
        end

        local rect = SDL.newRect()
        for y = 0, 14 do
            for x = 0, 19 do
                if x > 0 and x < 19 and y > 0 and y < 13 then
                    if map:monster_allowed(x, y) then
                        local t  = not map:monster_allowed(x    , y - 1) -- true if grass to north of this cell
                        local tr = not map:monster_allowed(x + 1, y - 1)
                        local r  = not map:monster_allowed(x + 1, y    )
                        local br = not map:monster_allowed(x + 1, y + 1)
                        local b  = not map:monster_allowed(x    , y + 1)
                        local bl = not map:monster_allowed(x - 1, y + 1)
                        local l  = not map:monster_allowed(x - 1, y    )
                        local lt = not map:monster_allowed(x - 1, y - 1)
                        local tile
                        -- top left corner
                        SDL.setRect(rect, x * RECTSIZE_X, y * RECTSIZE_Y, RECTSIZE_X / 2, RECTSIZE_Y / 2)
                        if l and t then tile = 1           -- grass on left and top     L ? T = 1
                        elseif (not l) and t then tile = 2 -- grass on top              - ? T = 2
                        elseif l and (not t) then tile = 4 -- grass on left             L ? - = 4
                        elseif lt then tile = 13           -- grass in left-top corner  - C - = 13
                        else tile = 5                      -- all path                  - - - = 5
                        end
                        SDL.BlitSurface(tiles, level_tiles[tile], bg, rect)
                        -- top right corner
                        SDL.setRect(rect, x * RECTSIZE_X + RECTSIZE_X / 2, y * RECTSIZE_Y, RECTSIZE_X / 2, RECTSIZE_Y / 2)
                        if t and r then tile = 3           -- T ? R = 3 
                        elseif (not t) and r then tile = 6 -- - ? R = 6 
                        elseif t and (not r) then tile = 2 -- T ? - = 2 
                        elseif tr then tile = 12           -- - C - = 12
                        else tile = 5                      -- - - - = 5 
                        end
                        SDL.BlitSurface(tiles, level_tiles[tile], bg, rect)
                        -- bottom left corner
                        SDL.setRect(rect, x * RECTSIZE_X, y * RECTSIZE_Y + RECTSIZE_Y / 2, RECTSIZE_X / 2, RECTSIZE_Y / 2)
                        if b and l then tile = 7           -- B ? L = 7 
                        elseif (not b) and l then tile = 4 -- - ? L = 4 
                        elseif b and (not l) then tile = 8 -- B ? - = 8 
                        elseif bl then tile = 11           -- - C - = 11
                        else tile = 5                      -- - - - = 5 
                        end
                        SDL.BlitSurface(tiles, level_tiles[tile], bg, rect)
                        -- bottom right corner
                        SDL.setRect(rect, x * RECTSIZE_X + RECTSIZE_X / 2, y * RECTSIZE_Y + RECTSIZE_Y / 2, RECTSIZE_X / 2, RECTSIZE_Y / 2)
                        if b and r then tile = 9           -- B ? R = 9 
                        elseif (not b) and r then tile = 6 -- - ? R = 6 
                        elseif b and (not r) then tile = 8 -- B ? - = 8 
                        elseif br then tile = 10           -- - C - = 10
                        else tile = 5                      -- - - - = 5 
                        end
                        SDL.BlitSurface(tiles, level_tiles[tile], bg, rect)
                    else
                        SDL.setRect(rect, x * RECTSIZE_X, y * RECTSIZE_Y, RECTSIZE_X, RECTSIZE_Y)
                        if map:tower_allowed(x, y) then
                            SDL.BlitSurface(tiles, level_tiles['g'], bg, rect)
                        else
                            SDL.BlitSurface(tiles, level_tiles['r'], bg, rect)
                        end
                    end
                else
                    SDL.setRect(rect, x * RECTSIZE_X, y * RECTSIZE_Y, RECTSIZE_X, RECTSIZE_Y)
                    SDL.BlitSurface(tiles, level_tiles['b'], bg, rect)
                end
            end
        end
-- SDL.SaveBMP(bg, "bg2.bmp") -- -FIX-
    end

    currentlevel = lid
end

