--  Awesome Tower Defense
--  Copyright (C) 2008-2009  Trygve Vea and contributors (read AUTHORS)
--  Copyright (C) 2010 Sean Bolton

--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.

--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.

--  You should have received a copy of the GNU General Public License along
--  with this program; if not, write to the Free Software Foundation, Inc.,
--  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

RECTSIZE_X  = 32
RECTSIZE_Y  = 32

SCREEN_MENU     = 1
SCREEN_INGAME   = 2
SCREEN_HIGHSCORES  = 3
SCREEN_CREDITS  = 4
SCREEN_TUTORIAL = 5

screen = nil         -- SDL_Surface *
background = nil     -- SDL_Surface *
menubackground = nil -- SDL_Surface *

-- An array of integers representing the playfield grid ( X x Y ) ...
-- If set to true, indicates that it should be updated. Usually set by
-- updaterect, which also makes sure that the background is blit first.
-- static int field[VIDEOMODE_WIDTH/RECTSIZE_X][VIDEOMODE_HEIGHT/RECTSIZE_Y];

-- Note: field and updatefield have been replace by the damage array object,
-- which is the same basic idea. See awesometd-ysdl.c for the implementation
-- and object methods.
damaged = nil

rendercount = 0
fpstext = ""
costtext = "   100"

current_screen = SCREEN_MENU

-- Some stuff we wanna do before we start drawing stuff.
function init_video()
    damaged = td.new_damage_array(VIDEOMODE_WIDTH, VIDEOMODE_HEIGHT, RECTSIZE_X, RECTSIZE_Y)

    screen = SDL.SetVideoMode(VIDEOMODE_WIDTH, VIDEOMODE_HEIGHT, VIDEOMODE_DEPTH, SDL.HWSURFACE)
    SDL.WM_SetCaption("Awesome Tower Defense 0.3","")
    background = SDL.CreateRGBSurface(SDL.HWSURFACE, VIDEOMODE_WIDTH, VIDEOMODE_HEIGHT, VIDEOMODE_DEPTH, 0, 0, 0, 0)
    background = background:DisplayFormat()
    menubackground = load_surface_from_file("menuscreen.bmp"):DisplayFormat()

    init_sprites()
    init_buttons()

    -- background:FillRect(nil, SDL.MapRGB_using_Surface(background,0,0,0))
    load_level(background, currentlevel)
    update_all()

    return 0
end


-- DRAW STUFF!
function render()
    draw_stuff_on_top()

    if damaged:has_damage() then
        damaged:do_UpdateRects(screen)
    end
    rendercount = rendercount + 1
end

-- Reports FPS
function getrenders()
    local i = rendercount
    fpstext = string.format("%11d fps", i)
    updateextent(520, 0, 120, 8)
    rendercount = 0
    return i
end

-- Marks a cell in the grid as one who needs to be updated.
-- This function should be called before we draw any sprites in the cell, since
-- it will blit from the background surface if we have one.
-- void updaterect(int x, int y) {
function updaterect(x, y)
    if x < 20 and y < 15 then
        if not damaged:get(x, y) then
            damaged:set(x, y, true)
            local rect = SDL.newRect(x * RECTSIZE_X, y * RECTSIZE_Y, RECTSIZE_X, RECTSIZE_Y)
            if current_screen == SCREEN_INGAME then
                SDL.BlitSurface(background, rect, screen, rect)
            elseif current_screen == SCREEN_MENU or current_screen == SCREEN_CREDITS or current_screen == SCREEN_HIGHSCORES or current_screen == SCREEN_TUTORIAL then
                SDL.BlitSurface(menubackground, rect, screen, rect)
            end
        end
    end
end

function updateextent(x, y, w, h)
    for ay = math.floor(y / RECTSIZE_Y), math.floor((y + h - 1) / RECTSIZE_Y) do
        for ax = math.floor(x / RECTSIZE_X), math.floor((x + w - 1) / RECTSIZE_X) do
            updaterect(ax, ay)
        end
    end
end

function update_all()
    if current_screen == SCREEN_INGAME then
        SDL.BlitSurface(background, nil, screen, nil)
    else
        SDL.BlitSurface(menubackground, nil, screen, nil)
    end
    for x = 0, 19 do
        for y = 0, 14 do
            updaterect(x,y)
        end
    end
end

function draw_stuff_on_top()

    if current_screen == SCREEN_CREDITS then
        draw_credits()
    elseif current_screen == SCREEN_HIGHSCORES then
        draw_highscores()
    elseif current_screen == SCREEN_TUTORIAL then
        draw_tutorial()
    elseif current_screen == SCREEN_MENU then
        if damaged:get(0, 6) or  -- -FIX- this first one is for the level-section menu hack
           damaged:get(12, 6) or damaged:get(13, 6) or damaged:get(14, 6) or damaged:get(15, 6) or
           damaged:get(12, 7) or damaged:get(13, 7) or damaged:get(14, 7) or damaged:get(15, 7) or
           damaged:get(12, 8) or damaged:get(13, 8) or damaged:get(14, 8) or damaged:get(15, 8) then
            draw_menu()
        end

    elseif current_screen == SCREEN_INGAME then
        for x = 0, 19 do
            for y = 0, 13 do
                if damaged:get(x, y) then
                    -- -FIX- this is horribly inefficient, and would result in half-drawn enemies if they weren't constantly moving
                    draw_tower(x,y)
                    draw_enemy(x,y)
                    draw_projectile(x,y)
                end
            end
        end
        if damaged:get(cell_x, cell_y) then draw_cursor(screen) end

        -- erase (any portion of) monster sprites that may be in the borders
        local rect = SDL.newRect()
        for _, start in pairs(map:get_starts()) do
            if start.x == 0 or start.x == 19 or start.y == 0 or start.y >= 13 then
                updaterect(start.x, start.y)
                SDL.setRect(rect, start.x * RECTSIZE_X, start.y * RECTSIZE_Y, RECTSIZE_X, RECTSIZE_Y)
                screen:FillRect(rect, SDL.MapRGB_using_Surface(screen,0,0,0))
            end
        end
        for _, exit in pairs(map:get_exits()) do
            if exit.x == 0 or exit.x == 19 or exit.y == 0 or exit.y >= 13 then
                updaterect(exit.x, exit.y)
                SDL.setRect(rect, exit.x * RECTSIZE_X, exit.y * RECTSIZE_Y, RECTSIZE_X, RECTSIZE_Y)
                screen:FillRect(rect, SDL.MapRGB_using_Surface(screen,0,0,0))
            end
        end

        --
        if damaged:get(8, 13) or damaged:get(9, 13) or damaged:get(10, 13) or damaged:get(11, 13) or damaged:get(12, 13) or
           damaged:get(8, 14) or damaged:get(9, 14) or damaged:get(10, 14) or damaged:get(11, 14) or damaged:get(12, 14) then
            draw_buttons(screen)
        end
        if damaged:get(8, 13) or damaged:get(9, 13) then
            draw_text_color(screen,costtext,265,416,255,255,0)
        end
        if damaged:get(10, 13) or damaged:get(11, 13) or damaged:get(12, 13) or damaged:get(13, 13) or damaged:get(14, 13) or damaged:get(15, 13) then
            draw_hint_text(screen)
        end
        if damaged:get(3, 13) or damaged:get(4, 13) or damaged:get(5, 13) or
           damaged:get(3, 14) or damaged:get(4, 14) or damaged:get(5, 14) then
            draw_numbers()
        end
        if damaged:get(16, 0) or damaged:get(17, 0) or damaged:get(18, 0) or damaged:get(19, 0) then
            draw_text(screen, fpstext, 520, 0)
        end

        if damaged:get(8, 0) or damaged:get(9, 0) or damaged:get(10, 0) or damaged:get(11, 0) or damaged:get(12, 0) or damaged:get(13, 0) or damaged:get(14, 0) or damaged:get(15, 0) then
            draw_buttons2(screen)
        end

    end
end

