--  Awesome Tower Defense
--  Copyright (C) 2008-2009  Trygve Vea and contributors (read AUTHORS)
--  Copyright (C) 2010 Sean Bolton
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License along
--  with this program; if not, write to the Free Software Foundation, Inc.,
--  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

MAX_DIRECTIONS = 8
MAX_MONSTERS = 50
MAX_TOWERS = 100
MAX_PROJECTILES = 128

DIRECTION_N  = 0
DIRECTION_NE = 1
DIRECTION_E  = 2
DIRECTION_SE = 3
DIRECTION_S  = 4
DIRECTION_SW = 5
DIRECTION_W  = 6
DIRECTION_NW = 7

ALGORITHM_TRAVELLED_SHORTEST = 1
ALGORITHM_TRAVELLED_FARTHEST = 2
ALGORITHM_MOST_HP = 3
ALGORITHM_LEAST_HP = 4
ALGORITHM_FASTEST = 5

-- struct monster {
--     int pos_x, pos_y; // Exact coordinates, not the cell-coordinates...
--                     // We need the exact ones, because we want smooth movements.
--     int dest_x, dest_y;
--     int spid; // Sprite ID
--     int frameno;
--     int cur_hp, max_hp;
--     int upcoming_damage;
--     int speed;
--     int direction;
--     int progress;
--     int travelled;
--     int score;
--     int money;
--     Uint32 last_ice_shot;
--     float effect_speed;
--     float upcoming_effect_speed;
--     int exp_gainer;
-- };

-- struct tower {
--     int loc_x, loc_y; // The grid location ... these doesn't move around.
--     int spid; // Sprite ID
--     int frameno;
--     int active;  -- boolean
--     int reload;
--     int damage;
--     int splash;
--     int price;
--     int reloadtimeleft;
--     int range;
--     int projectile;
--     int target_algorithm;
--     int exp;
--     int max_exp;
--     int next_tower;
--     int paused;  -- boolean
-- };

-- struct sprites {
--     SDL_Surface *sprites[MAX_SPRITES];
--     int frames[MAX_SPRITES];
--     SDL_Rect slices[MAX_SPRITES][MAX_DIRECTIONS][MAX_FRAMES];
--     int spritecount;
-- };

-- struct projectile {
--     float pos_x, pos_y; // The exact location.
--     int spid; // Sprite ID
--     int frameno;
--     int speed;
--     int targetmonster;
--     int damage; // If damage is 0, it's considered as inactive.
--     int splash;
--     float effect_speed;
--     int owner;
--     Uint32 color;
-- };

local function get_locx(__A) return math.floor(__A/RECTSIZE_X) end
local function get_locy(__A) return math.floor(__A/RECTSIZE_Y) end
local function get_cellx(__A) return math.floor(__A.pos_x/RECTSIZE_X) end
local function get_celly(__A) return math.floor(__A.pos_y/RECTSIZE_Y) end


-- static const int level_monster[1][25][40] = {  -- indexed from 1
level_monster = {
    {
        { 1,1,1,1,1,1,1,1,1,1 },
        { 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2 },
        { 3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3 },
        { 4,4,4,4,4,4,4,4,4,4,4,4,4,4,4 },
        { 5,5,5,5,5,5,5,5,5,5 },
        { 6,6,6,6,6,6,6,6,6,6,6,6,6,6,6 },
        { 7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7 },
        { 8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8 },
        { 9,9,9,9,9,9,9,9,9,9,9,9,9,9,9 },
        { 10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10 },
        { 11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11 },
        { 12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12 },
        { 13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13 },
        { 14,14,14,14,14,14,14,14,14,14,14,14,14,14,14 },
        { 15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15 },
        { 16,16,16,16,16,16,16,16,16,16,16,16,16,16,16 },
        { 17,17,17,17,17,17,17,17,17,17,17,17,17,17,17 },
        { 18,18,18,18,18,18,18,18,18,18,18,18,18,18,18 },
        { 19,19,19,19,19,19,19,19,19,19,19,19,19,19,19 },
        { 20,20,20,20,20,20,20,20,20,20,20,20,20,20,20 },
        { 21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21 },
        { 22,22,22,22,22,22,22,22,22,22,22,22,22,22,22 },
        { 23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23 },
        { 24,24,24,24,24,24,24,24,24,24,24,24,24,24,24 },
        { 25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25 }
    }
}

-- static const int monster_batches[1][25] = {  -- indexed from 1
monster_batches = {
    { 
        10, 15, 30, 15, 10, 15, 20, 20, 15, 20,
        20, 25, 20, 15, 20, 15, 15, 15, 15, 15,
        20, 15, 20, 15, 20
    }
}

-- static const int level_batches[1] = {  -- indexed from 1
level_batches = {
    25
}

-- static const struct monster monster_definitions[25] = {  -- indexed from 1
monster_definitions = {
    { 0,0,0,0,0,0,  15,  15,0,4,0,0,0,  10, 1,0,1,1,-1 },
    { 0,0,0,0,4,0,  20,  20,0,5,0,0,0,  20, 2,0,1,1,-1 },
    { 0,0,0,0,0,0,  50,  50,0,4,0,0,0,  35, 3,0,1,1,-1 },
    { 0,0,0,0,0,0, 100, 100,0,3,0,0,0,  50, 4,0,1,1,-1 },
    { 0,0,0,0,4,0, 150, 150,0,5,0,0,0, 100, 5,0,1,1,-1 },
    { 0,0,0,0,0,0, 200, 200,0,6,0,0,0, 150, 6,0,1,1,-1 },
    { 0,0,0,0,2,0, 300, 300,0,7,0,0,0, 200, 7,0,1,1,-1 },
    { 0,0,0,0,0,0, 400, 400,0,6,0,0,0, 250, 8,0,1,1,-1 },
    { 0,0,0,0,0,0, 600, 600,0,6,0,0,0, 300, 9,0,1,1,-1 },
    { 0,0,0,0,0,0, 700, 700,0,6,0,0,0, 350,10,0,1,1,-1 },
    { 0,0,0,0,0,0, 800, 800,0,6,0,0,0, 400,11,0,1,1,-1 },
    { 0,0,0,0,0,0,1000,1000,0,5,0,0,0, 450,12,0,1,1,-1 },
    { 0,0,0,0,0,0,1200,1200,0,6,0,0,0, 500,13,0,1,1,-1 },
    { 0,0,0,0,0,0,1400,1400,0,7,0,0,0, 550,14,0,1,1,-1 },
    { 0,0,0,0,0,0,1600,1600,0,6,0,0,0, 600,15,0,1,1,-1 },
    { 0,0,0,0,0,0,1800,1800,0,6,0,0,0, 650,16,0,1,1,-1 },
    { 0,0,0,0,0,0,2000,2000,0,8,0,0,0, 700,17,0,1,1,-1 },
    { 0,0,0,0,0,0,2250,2250,0,7,0,0,0, 750,18,0,1,1,-1 },
    { 0,0,0,0,0,0,2500,2500,0,7,0,0,0, 800,19,0,1,1,-1 },
    { 0,0,0,0,0,0,2750,2750,0,7,0,0,0, 850,20,0,1,1,-1 },
    { 0,0,0,0,0,0,3000,3000,0,7,0,0,0, 900,21,0,1,1,-1 },
    { 0,0,0,0,0,0,3250,3250,0,7,0,0,0, 950,22,0,1,1,-1 },
    { 0,0,0,0,0,0,3500,3500,0,7,0,0,0,1000,23,0,1,1,-1 },
    { 0,0,0,0,0,0,3750,3750,0,7,0,0,0,1100,24,0,1,1,-1 },
    { 0,0,0,0,0,0,4000,4000,0,7,0,0,0,1250,25,0,1,1,-1 }
}

-- const struct tower tower_definitions[9] = {  -- indexed from 1
tower_definitions = {
    { 0,0,1,0,1,5,0,0,100,0,80,0,ALGORITHM_TRAVELLED_FARTHEST,0,10,1,0 },
    { 0,0,6,0,1,4,0,0,200,0,125,1,ALGORITHM_TRAVELLED_FARTHEST,0,20,2,0 },
    { 0,0,9,0,1,3,0,0,400,0,200,2,ALGORITHM_TRAVELLED_FARTHEST,0,0,0,0 }, 
    { 0,0,3,0,1,40,0,0,300,0,75,3,ALGORITHM_TRAVELLED_FARTHEST,0,10,4,0 },
    { 0,0,7,0,1,30,0,0,600,0,90,4,ALGORITHM_TRAVELLED_FARTHEST,0,20,5,0 },
    { 0,0,10,0,1,20,0,0,900,0,130,5,ALGORITHM_TRAVELLED_FARTHEST,0,0,0,0 },
    { 0,0,5,0,1,10,0,0,150,0,100,6,ALGORITHM_FASTEST,0,10,7,0 },
    { 0,0,8,0,1,7,0,0,300,0,125,7,ALGORITHM_FASTEST,0,20,8,0 },
    { 0,0,11,0,1,5,0,0,450,0,150,8,ALGORITHM_FASTEST,0,0,0,0 }
}

-- static struct projectile projectile_definitions[9] = {  -- indexed from 1
projectile_definitions = {
    { 0,0,1,0, 7,0, 5, 0, 0,0,0 },
    { 0,0,1,0, 7,0, 10, 0, 0,0,0 },
    { 0,0,1,0, 7,0, 25, 0, 0,0,0 },
    { 0,0,1,0, 1,0, 90, 50, 0,0,0 },
    { 0,0,1,0, 2,0, 180, 50, 0,0,0 },
    { 0,0,1,0, 2,0, 360, 50, 0,0,0 },
    { 0,0,1,0, 3,0, 5, 0, -0.25,0,0 },
    { 0,0,1,0, 3,0, 7, 0, -0.30,0,0 },
    { 0,0,1,0, 3,0, 10, 0, -0.40,0,0 }
}

level = 0
batch = 0
monster = 0

selected_tower = 0

-- struct monster monsters[MAX_MONSTERS];  -- indexed from 0
monsters = {}
-- struct tower towers[MAX_TOWERS];  -- indexed from 1
towers = {}
-- struct projectile projectiles[MAX_PROJECTILES];  -- indexed from 0
projectiles = {}
-- struct sprites sprites;
sprites = { sprites = {}, frames = {}, slices = {}, spritecount = 0 }
gfx_pause = nil -- SDL_Surface *

function reset_everything()
    level = 0
    batch = 0
    monster = 0
    selected_tower = 0
    -- memset(&monsters, 0x00, sizeof(struct monster)*MAX_MONSTERS);
    for i = 0, MAX_MONSTERS - 1 do
        monsters[i] = {}
        monsters[i].cur_hp = 0
    end
    -- memset(&towers, 0x00, sizeof(struct tower)*MAX_TOWERS);
    for i = 1, MAX_TOWERS do
        towers[i] = {}
        towers[i].active = false
    end
    -- memset(&projectiles, 0x00, sizeof(struct projectile)*MAX_PROJECTILES);
    for i = 0, MAX_PROJECTILES - 1 do
        projectiles[i] = {}
        projectiles[i].damage = 0
    end
    map:reset()
end

function toggle_shooting(tower)
    towers[tower].paused = not towers[tower].paused
    updaterect(14,0)
    updaterect(towers[tower].loc_x,towers[tower].loc_y)
--    printf("toggling for %d: (currently: %d)\n", tower, towers[tower].paused);
end

function load_surface_from_file(filename)
    local surface = IMG.Load("./data/" .. filename)
    if not surface then
        print("Error: could not load image from file:", SDL.GetError())
        return nil
    end
    return surface
end

-- void load_sprite_from_pic(char *filename, int width, int height, int frames, int memsize) {
function load_sprite_from_pic(filename, width, height, frames)
    sprite = load_surface_from_file(filename)
    if sprite then
        sprites.frames[sprites.spritecount] = frames
        -- hack to get color keying to work on SDL 1.2.14 Quartz backend on Mac OS X Tiger Intel !FIX!
        sprite = sprite:DisplayFormat()
        local colorkey = SDL.MapRGB_using_Surface(sprite, 255,0,255)
        if colorkey > 0xffffff then
          colorkey = colorkey + 255
        end
        -- end hack
        sprite:SetColorKey(SDL.SRCCOLORKEY, colorkey)
        -- sprite = sprite:DisplayFormat()  !FIX! re-enable if hack not needed
        sprites.slices[sprites.spritecount] = {}
        local w, h = sprite:getSize()
        for d = 0, MAX_DIRECTIONS - 1 do
            local di
            if h == height * 8 then -- spritesheet has diagonals
                di = d
            else -- old spritesheet without diagonals
                di = math.floor(d / 2)
            end
            sprites.slices[sprites.spritecount][d] = {}
            for i = 0, frames - 1 do
                sprites.slices[sprites.spritecount][d][i] =
                    SDL.newRect(0 + (i*width), 0 + (di*height), width, height)
            end
        end
        sprites.sprites[sprites.spritecount] = sprite
        print(string.format("* Finished adding sprite %d!", sprites.spritecount))
        sprites.spritecount = sprites.spritecount + 1
    else
        print(string.format("Big bad error while loading file: \"%s\": %s", filename, SDL.GetError()))
    end
end

-- void init_sprites(void) {
function init_sprites()
    load_sprite_from_pic("enemy1.bmp",32,32,2)
    load_sprite_from_pic("tower1.bmp",32,32,1)
    load_sprite_from_pic("enemy2.bmp",32,32,2)
    load_sprite_from_pic("tower2.bmp",32,32,8)
    load_sprite_from_pic("enemy3.bmp",32,32,2)
    load_sprite_from_pic("tower3.bmp",32,32,1)
    load_sprite_from_pic("tower1-2.bmp",32,32,1)
    load_sprite_from_pic("tower2-2.bmp",32,32,8)
    load_sprite_from_pic("tower3-2.bmp",32,32,1)
    load_sprite_from_pic("tower1-3.bmp",32,32,1)
    load_sprite_from_pic("tower2-3.bmp",32,32,8)
    load_sprite_from_pic("tower3-3.bmp",32,32,1)
    gfx_pause = load_surface_from_file("minipause.bmp")
    gfx_pause:SetColorKey(SDL.SRCCOLORKEY, SDL.MapRGB_using_Surface(gfx_pause, 255,0,255)) -- !FIX! doesn't have Tiger/Quartz hack
    gfx_pause = gfx_pause:DisplayFormat()

    projectile_definitions[1][11] = SDL.MapRGB_using_Surface(screen, 128, 128, 128)
    projectile_definitions[2][11] = SDL.MapRGB_using_Surface(screen, 192, 192, 192)
    projectile_definitions[3][11] = SDL.MapRGB_using_Surface(screen, 255, 255, 255)
    projectile_definitions[4][11] = SDL.MapRGB_using_Surface(screen, 128, 0, 0)
    projectile_definitions[5][11] = SDL.MapRGB_using_Surface(screen, 192, 0, 0)
    projectile_definitions[6][11] = SDL.MapRGB_using_Surface(screen, 255, 0, 0)
    projectile_definitions[7][11] = SDL.MapRGB_using_Surface(screen, 0, 0, 128)
    projectile_definitions[8][11] = SDL.MapRGB_using_Surface(screen, 0, 0, 192)
    projectile_definitions[9][11] = SDL.MapRGB_using_Surface(screen, 0, 0, 255)
end

-- void draw_sprite(SDL_Surface *s, int spid, int fid, int rot, int x, int y) {
function draw_sprite(s, spid, fid, rot, x, y)
    local dst = SDL.newRect()
    if spid < sprites.spritecount then
        if fid < sprites.frames[spid] then
            if x < VIDEOMODE_WIDTH and y < VIDEOMODE_HEIGHT then
                local _, _, w, h = SDL.getRect(sprites.slices[spid][rot][fid])
                SDL.setRect(dst, x, y, w, h)
                SDL.BlitSurface(sprites.sprites[spid], sprites.slices[spid][rot][fid], s, dst)
            else
                print("ERR -- draw_sprite(): Attempt to draw outside the screen.")
            end
        else
            print("ERR -- draw_sprite(): Frame doesn't exist.")
        end
    else
        print("ERR -- draw_sprite(): Sprite doesn't exist.", spid)
    end
end

function select_tower(x, y)
    local n = map:get_tower(x, y)
    if n ~= 0 then
        selected_tower = n
        updateextent(256, 8, 216, 24)
    end
end

function sell_tower(tid)
    towers[tid].active = false
    map:put_tower(towers[tid].loc_x, towers[tid].loc_y, 0)
    if levels[currentlevel].pathfinding == "lee" then
        redest_monsters()
    end
    update_money(math.floor(towers[tid].price/2))
    selected_tower = 0
    updaterect(towers[tid].loc_x, towers[tid].loc_y)
    updateextent(256, 8, 216, 24)
end

function tower_algorithm(tid, aid)
    updateextent(304, 8, 120, 24)
    towers[tid].target_algorithm = aid
end

function set_monster_dest(m)
    local newdir = map:monster_direction(m.pos_x, m.pos_y, m.direction)
    local cell_x = math.floor(m.pos_x / RECTSIZE_X)
    local cell_y = math.floor(m.pos_y / RECTSIZE_Y)
    if newdir == DIRECTION_N then
        cell_y = cell_y - 1
    elseif newdir == DIRECTION_NE then
        cell_x = cell_x + 1
        cell_y = cell_y - 1
    elseif newdir == DIRECTION_E then
        cell_x = cell_x + 1
    elseif newdir == DIRECTION_SE then
        cell_x = cell_x + 1
        cell_y = cell_y + 1
    elseif newdir == DIRECTION_S then
        cell_y = cell_y + 1
    elseif newdir == DIRECTION_SW then
        cell_x = cell_x - 1
        cell_y = cell_y + 1
    elseif newdir == DIRECTION_W then
        cell_x = cell_x - 1
    elseif newdir == DIRECTION_NW then
        cell_x = cell_x - 1
        cell_y = cell_y - 1
    end
    m.direction = newdir
    m.dest_x = cell_x * RECTSIZE_X + RECTSIZE_X / 2
    m.dest_y = cell_y * RECTSIZE_Y + RECTSIZE_Y / 2
end

function redest_monsters()
    for i = 0, MAX_MONSTERS - 1 do
        local m = monsters[i]
        if m.cur_hp > 0 and not map:is_exit(get_cellx(m), get_celly(m)) then
            set_monster_dest(m)
        end
    end
end

-- int spawn_monster(void) {
function spawn_monster()
    if monster_batches[level + 1][batch + 1] == monster then
        for i = 0, MAX_MONSTERS - 1 do
            if monsters[i].cur_hp > 0 then
--              print("Not spawning: Stuff is still alive.")
--   Not spawning any new monsters, 
--   because we still have live monsters on the field.
                return false
            end
        end
        monster = 0
        batch = batch + 1
        if batch == level_batches[level + 1] then
            show_highscores(SCORE_WONMESSAGE)
            return false
        end
        update_batch()
        return false
    end

    for i = 0, MAX_MONSTERS - 1 do
        if monsters[i].cur_hp == 0 then
            local t = monster_definitions[level_monster[level + 1][batch + 1][monster + 1]]
            --
            local m = monsters[i]
            m.spid      = t[5]
            m.frameno   = t[6]
            m.cur_hp    = t[7]
            m.max_hp    = t[8]
            m.upcoming_damage = t[9]
            m.speed     = t[10]
            m.direction = t[11]
            m.progress  = t[12]
            m.travelled = t[13]
            m.score     = t[14]
            m.money     = t[15]
            m.last_ice_shot = t[16]
            m.effect_speed = t[17]
            m.upcoming_effect_speed = t[18]
            m.exp_gainer = t[19]
            --
            local starts = map:get_starts()
            local start = math.random(#starts)
            m.pos_x = starts[start].x * RECTSIZE_X + RECTSIZE_X / 2
            m.pos_y = starts[start].y * RECTSIZE_Y + RECTSIZE_Y / 2
            set_monster_dest(m)
            monster = monster + 1
            return false
        end
    end
    return false
end

function move_projectile()
    local ph, pw = 4, 4
    local hph, hpw = 2, 2

    for i = 0, MAX_PROJECTILES - 1 do
        if projectiles[i].damage > 0 and monsters[projectiles[i].targetmonster].cur_hp <= 0 then
            updateextent(projectiles[i].pos_x, projectiles[i].pos_y, pw, ph)
            projectiles[i].damage = 0

        elseif projectiles[i].damage > 0 then
            local targetmonster = monsters[projectiles[i].targetmonster]

--             /*angle = atan2(targetmonster.pos_y+16 - projectiles[i].pos_y, targetmonster.pos_x+16 - projectiles[i].pos_x);
--             x = cos(angle);
--             y = sin(angle);*/

            -- Calculate direction without trigonometric functions. Should be faster
            local dx = targetmonster.pos_x - projectiles[i].pos_x
            local dy = targetmonster.pos_y - projectiles[i].pos_y
            local d = math.sqrt(dx*dx + dy*dy)
            local x = dx/d
            local y = dy/d

            updateextent(projectiles[i].pos_x, projectiles[i].pos_y, pw, ph)

            projectiles[i].pos_x = projectiles[i].pos_x + x * projectiles[i].speed
            projectiles[i].pos_y = projectiles[i].pos_y + y * projectiles[i].speed

            updateextent(projectiles[i].pos_x, projectiles[i].pos_y, pw, ph)

            if math.abs((projectiles[i].pos_x + hpw) - targetmonster.pos_x) < 16 and
               math.abs((projectiles[i].pos_y + hph) - targetmonster.pos_y) < 16 then
                targetmonster.cur_hp = targetmonster.cur_hp - projectiles[i].damage
                if towers[projectiles[i].owner].max_exp > 0 then targetmonster.exp_gainer = projectiles[i].owner end

                if projectiles[i].splash > 0 then
                    for m = 0, MAX_MONSTERS - 1 do
                        if monsters[m].cur_hp > 0 and m ~= projectiles[i].targetmonster then
                            local k1 = (projectiles[i].pos_x+hpw)-(monsters[m].pos_x)
                            local k2 = (projectiles[i].pos_y+hph)-(monsters[m].pos_y)

                            local h = (k1*k1)+(k2*k2)
                            if h <= (projectiles[i].splash*projectiles[i].splash) then
                                local dmg = 0.5 * projectiles[i].damage
                                local modifier = projectiles[i].splash*projectiles[i].splash/h
                                local finaldmg = dmg * modifier
                                monsters[m].cur_hp = monsters[m].cur_hp - finaldmg
                                if towers[projectiles[i].owner].max_exp > 0 then monsters[m].exp_gainer = projectiles[i].owner end

                                if monsters[m].cur_hp <= 0 then
                                    monsters[m].cur_hp = 0;
                                    updateextent(monsters[m].pos_x - 16, monsters[m].pos_y - 16, RECTSIZE_X, RECTSIZE_Y)
                                    -- rewarding the player for early kills.
                                    update_score(monsters[m].score-(monsters[m].score*(monsters[m].travelled/1728)))
                                    if monsters[m].exp_gainer > -1 then
                                        if towers[monsters[m].exp_gainer].next_tower > 0 then towers[monsters[m].exp_gainer].exp = towers[monsters[m].exp_gainer].exp + 1 end
                                        if towers[monsters[m].exp_gainer].exp >= towers[monsters[m].exp_gainer].max_exp and towers[monsters[m].exp_gainer].next_tower > 0 then
                                            upgrade_tower(monsters[m].exp_gainer)
                                        end
                                    end
                                    update_money(monsters[m].money)
                                end
                            end
                        end
                    end
                end


                if projectiles[i].effect_speed < 1 then
                    if targetmonster.effect_speed == 1 then
                        targetmonster.effect_speed = targetmonster.effect_speed + projectiles[i].effect_speed
                    end
                    targetmonster.last_ice_shot = gamecycle
                end
                targetmonster.upcoming_damage = targetmonster.upcoming_damage - projectiles[i].damage
                targetmonster.upcoming_effect_speed = targetmonster.upcoming_effect_speed - projectiles[i].effect_speed
                projectiles[i].damage = 0
                if targetmonster.cur_hp <= 0 then
                    targetmonster.cur_hp = 0

                    updateextent(targetmonster.pos_x - 16, targetmonster.pos_y - 16, RECTSIZE_X, RECTSIZE_Y)

                    update_score(targetmonster.score-(targetmonster.score*(targetmonster.travelled/1728)))
                    if targetmonster.exp_gainer > -1 then
                        if towers[targetmonster.exp_gainer].next_tower > 0 then towers[targetmonster.exp_gainer].exp = towers[targetmonster.exp_gainer].exp + 1 end
                        if towers[targetmonster.exp_gainer].exp >= towers[targetmonster.exp_gainer].max_exp and towers[targetmonster.exp_gainer].next_tower > 0 then
                            upgrade_tower(targetmonster.exp_gainer)
                        end
                    end
                    update_money(targetmonster.money)
                end
            end
        end
    end
end

function move_monster()
    for i = 0, MAX_MONSTERS - 1 do
        local m = monsters[i]
        if m.cur_hp > 0 then
            updateextent(m.pos_x - 16, m.pos_y - 16, 32, 32)

            m.progress = m.progress + (m.speed * m.effect_speed)
            if (m.last_ice_shot + 300) < gamecycle then m.effect_speed = 1 end

            while m.progress > 10 do
                m.travelled = m.travelled + 1
                local dx = math.abs(m.pos_x - m.dest_x)
                local dy = math.abs(m.pos_y - m.dest_y)
                if dx == dy then
                    if m.pos_x < m.dest_x then m.pos_x = m.pos_x + 1
                    elseif m.pos_x > m.dest_x then m.pos_x = m.pos_x - 1 end
                    if m.pos_y < m.dest_y then m.pos_y = m.pos_y + 1
                    elseif m.pos_y > m.dest_y then m.pos_y = m.pos_y - 1 end
                elseif dx > dy then
                    if m.pos_x < m.dest_x then m.pos_x = m.pos_x + 1
                    elseif m.pos_x > m.dest_x then m.pos_x = m.pos_x - 1 end
                else
                    if m.pos_y < m.dest_y then m.pos_y = m.pos_y + 1
                    elseif m.pos_y > m.dest_y then m.pos_y = m.pos_y - 1 end
                end
                if m.pos_x == m.dest_x and m.pos_y == m.dest_y then
                    if map:is_exit(get_cellx(m), get_celly(m)) then
                        -- monster has arrived at the end of its path
                        update_lives(-1)
                        if current_screen ~= SCREEN_INGAME then return end
                        m.cur_hp = 0
                        break
                    end
                    set_monster_dest(m)
                end
                m.progress = m.progress - 10
            end
            if m.cur_hp > 0 then
                updateextent(m.pos_x - 16, m.pos_y - 16, 32, 32)
            end
        end
    end
end

-- void draw_health(SDL_Surface *s, int x, int y, int cur, int max) {
function draw_health(s, x, y, cur, max)
    local green_w = math.floor(28 * cur / max)
    local green = SDL.newRect( x+2,y,green_w,3 )
    local red = SDL.newRect( x+2+green_w,y,28-green_w,3 )
    s:FillRect(green, SDL.MapRGB_using_Surface(s,0,255,0))
    s:FillRect(red, SDL.MapRGB_using_Surface(s, 255,0,0))
end
-- void draw_reload(SDL_Surface *s, int x, int y, int cur, int max) {
function draw_reload(s, x, y, cur, max)
    local gray = SDL.newRect( x+2,y+28,(28*cur/max),3 )
    s:FillRect(gray, SDL.MapRGB_using_Surface(s,200,200,200))
end
-- void draw_exp(SDL_Surface *s, int x, int y, int cur, int max) {
function draw_exp(s, x, y, cur, max)
    local yellow = SDL.newRect( x+2,y+25,(28*cur/max),3 )
    s:FillRect(yellow, SDL.MapRGB_using_Surface(s,255,255,0))
end

-- void draw_enemy(int x, int y) {
function draw_enemy(x, y)
    for i = 0, MAX_MONSTERS - 1 do
        if monsters[i].cur_hp > 0 and get_cellx(monsters[i]) == x and get_celly(monsters[i]) == y then
            draw_sprite(screen,monsters[i].spid, monsters[i].frameno,monsters[i].direction, monsters[i].pos_x - 16, monsters[i].pos_y - 16)
            draw_health(screen,monsters[i].pos_x - 16, monsters[i].pos_y - 16, monsters[i].cur_hp, monsters[i].max_hp)
        end
    end
end

-- void draw_projectile(int x, int y) {
function draw_projectile(x, y)
    for i = 0, MAX_PROJECTILES - 1 do
        if projectiles[i].damage > 0 and get_cellx(projectiles[i]) == x and get_celly(projectiles[i]) == y then
            local projrect = SDL.newRect( projectiles[i].pos_x, projectiles[i].pos_y, 4, 4 )
            screen:FillRect(projrect, projectiles[i].color)
        end
    end
end

-- void draw_tower(int x, int y) {
function draw_tower(x, y)
    for i = 1, MAX_TOWERS do
        if towers[i].active then
            if towers[i].loc_x == x and towers[i].loc_y == y then
                if selected_tower == i then
                    local trect = SDL.newRect( towers[i].loc_x*32, towers[i].loc_y*32,32,32 )
                    screen:FillRect( trect, SDL.MapRGB_using_Surface(screen, 0,0,255) )
                end
                draw_sprite(screen,towers[i].spid, towers[i].frameno, DIRECTION_N,towers[i].loc_x*RECTSIZE_X,towers[i].loc_y*RECTSIZE_Y)
                draw_reload(screen,towers[i].loc_x*RECTSIZE_X, towers[i].loc_y*RECTSIZE_Y, towers[i].reloadtimeleft, towers[i].reload)
                if towers[i].max_exp > 0 then
                    draw_exp(screen,towers[i].loc_x*RECTSIZE_X, towers[i].loc_y*RECTSIZE_Y, towers[i].exp, towers[i].max_exp)
                end
                if towers[i].paused then
                    local pause = SDL.newRect( (towers[i].loc_x*32)+24, (towers[i].loc_y*32)+24, 8, 8 )
                    SDL.BlitSurface(gfx_pause, nil, screen, pause);
                end
                return
            end
        end
    end
end

-- int has_tower(int x, int y) {
-- function has_tower(x, y)
--     for i = 1, MAX_TOWERS do
--         if towers[i].active then
--             if towers[i].loc_x == x and towers[i].loc_y == y then
--                 return true
--             end
--         end
--     end
--     return false
-- end

function has_monster(x, y)
    for i = 0, MAX_MONSTERS - 1 do
        local m = monsters[i]
        if m.cur_hp > 0 and
           math.floor(m.pos_x / RECTSIZE_X) == x and
           math.floor(m.pos_y / RECTSIZE_Y) == y then
            return true
        end
    end
    return false
end

function tower_definition(type)  -- smb: added function, unpacks tower_definitions[type]
  -- note that tower_definitions[] is indexed from 1, but this function uses type starting from 0
  local d = tower_definitions[type + 1]
  return {
    loc_x   = d[1],
    loc_y   = d[2],
    spid    = d[3],
    frameno = d[4],
    active  = (d[5] == 1),
    reload  = d[6],
    damage  = d[7],
    splash  = d[8],
    price   = d[9],
    reloadtimeleft = d[10],
    range   = d[11],
    projectile = d[12],
    target_algorithm = d[13],
    exp     = d[14],
    max_exp = d[15],
    next_tower = d[16],
    paused  = (d[17] == 1)
  }
end

function upgrade_tower(tid)
    local tmptower = towers[tid]
    towers[tid] = tower_definition(tmptower.next_tower)
    towers[tid].loc_x = tmptower.loc_x
    towers[tid].loc_y = tmptower.loc_y
    towers[tid].target_algorithm = tmptower.target_algorithm
end

-- void add_tower(int x, int y, int type) {
function add_tower(x, y, type)
    local tdef = tower_definition(type)
    if have_money(tdef.price) and
       map:get_tower(x, y) == 0 and
       not has_monster(x, y)   -- don't allow tower placement on top of monsters
       then
        local n = -1
        for i = 1, MAX_TOWERS do
            if not towers[i].active then
                n = i
                break
            end
        end
        if n ~= -1 and map:put_tower(x, y, n) then -- not blocking
            towers[n] = tdef
            towers[n].loc_x = x
            towers[n].loc_y = y
            update_money(tdef.price*-1)
            if levels[currentlevel].pathfinding == "lee" then
                redest_monsters()
            end
        else
            set_cursor_would_block(x, y)
        end
    end
end

function animate_sprites()
    for i = 0, MAX_MONSTERS - 1 do
        if monsters[i].cur_hp > 0 then
            if (monsters[i].frameno+1) < sprites.frames[monsters[i].spid] then monsters[i].frameno = monsters[i].frameno + 1
            else monsters[i].frameno = 0 end
        end
    end
    for i = 1, MAX_TOWERS do
        if towers[i].active then
            if (towers[i].frameno+1) < sprites.frames[towers[i].spid] then towers[i].frameno = towers[i].frameno + 1
            else towers[i].frameno = 0 end
            updaterect(towers[i].loc_x,towers[i].loc_y)
        end
    end
end

function projectile_definition(pid)  -- smb: added function, unpacks projectile_definitions[pid]
  -- note that projectile_definitions[] is indexed from 1, but this function uses pid starting from 0
  local d = projectile_definitions[pid + 1]
  return {
    pos_x   = d[1],
    pos_y   = d[2],
    spid    = d[3],
    frameno = d[4],
    speed   = d[5],
    targetmonster = d[6],
    damage  = d[7],
    splash  = d[8],
    effect_speed = d[9],
    owner   = d[10],
    color   = d[11]
  }
end

-- void add_projectile(int pid, int mid, int tid) {
function add_projectile(pid, mid, tid)
    for i = 0, MAX_PROJECTILES - 1 do
        if projectiles[i].damage == 0 then
            projectiles[i] = projectile_definition(pid)
            projectiles[i].pos_x = (towers[tid].loc_x * RECTSIZE_X) + (RECTSIZE_X / 2)
            projectiles[i].pos_y = (towers[tid].loc_y * RECTSIZE_Y) + (RECTSIZE_Y / 2)
            projectiles[i].targetmonster = mid
            projectiles[i].owner = tid
            monsters[mid].upcoming_damage = monsters[mid].upcoming_damage + projectiles[i].damage
            monsters[mid].upcoming_effect_speed = monsters[mid].upcoming_effect_speed + projectiles[i].effect_speed
            return
        end
    end
end

function shoot_towers()
    for i = 1, MAX_TOWERS do
        if towers[i].active and not towers[i].paused then
            if towers[i].reloadtimeleft == 0 then
                local target = -1
                for y = 0, MAX_MONSTERS - 1 do
                    if monsters[y].cur_hp > 0 and monsters[y].cur_hp > monsters[y].upcoming_damage then
                        local k1 = ((towers[i].loc_x*32)+16)-(monsters[y].pos_x)
                        local k2 = ((towers[i].loc_y*32)+16)-(monsters[y].pos_y)

                        local h = (k1*k1) + (k2*k2);

                        if h < towers[i].range*towers[i].range then
                            if target > -1 then
                                if towers[i].target_algorithm == ALGORITHM_MOST_HP then
                                    if monsters[y].cur_hp > monsters[target].cur_hp then target = y end
                                elseif towers[i].target_algorithm == ALGORITHM_LEAST_HP then
                                    if monsters[y].cur_hp < monsters[target].cur_hp then target = y end
                                elseif towers[i].target_algorithm == ALGORITHM_FASTEST then
                                    if (monsters[y].effect_speed*(monsters[y].speed*monsters[y].upcoming_effect_speed)) >
                                       (monsters[target].effect_speed*(monsters[target].speed*monsters[target].upcoming_effect_speed)) then
                                        target = y end
                                elseif towers[i].target_algorithm == ALGORITHM_TRAVELLED_SHORTEST then
                                    if monsters[y].travelled < monsters[target].travelled then target = y end
                                elseif towers[i].target_algorithm == ALGORITHM_TRAVELLED_FARTHEST then
                                    if monsters[y].travelled > monsters[target].travelled then target = y end
                                end
                            else
                                target = y
                            end
                        end
                    end
                end
                if target > -1 then
                    add_projectile(towers[i].projectile,target,i)
                    towers[i].reloadtimeleft = towers[i].reload
                    updaterect(towers[i].loc_x, towers[i].loc_y)
                end
            else
                towers[i].reloadtimeleft = towers[i].reloadtimeleft - 1
                updaterect(towers[i].loc_x, towers[i].loc_y)
            end
        end
    end
end

function update_batch()
    updateextent(96,450,96,8)
end
function draw_numbers2()
    local batchtext = string.format("batch  %2d/%2d", batch+1, level_batches[level + 1])
    draw_text(screen,batchtext,3*32,450)
end
